import json
import typing as tp
from abc import ABCMeta, abstractmethod

__all__ = ['JSONEncoder', 'JSONAble', 'json_encode']


class JSONAble(metaclass=ABCMeta):
    @abstractmethod
    def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
        """Return a JSON-able representation of this object"""


class JSONEncoder(json.JSONEncoder):
    def default(self, o) -> str:
        if hasattr(o, 'to_json'):
            return o.to_json()
        return super().default(o)


def json_encode(x) -> str:
    """
    Convert an object to JSON. Will properly handle subclasses of JSONAble

    :param x: object to convert
    """
    return JSONEncoder().encode(x)



import functools

__all__ = [
    'Singleton',
]


# Taken from https://wiki.python.org/moin/PythonDecoratorLibrary
def Singleton(cls):
    """
    Make a singleton out of decorated class.

    Usage:

    >>> @Singleton
    >>> class MyClass(object):
    >>>     ...
    """

    cls.__new_old__ = cls.__new__

    @functools.wraps(cls.__new__)
    def singleton_new(cls, *args, **kw):
        it = cls.__dict__.get('__it__')
        if it is not None:
            return it

        cls.__it__ = it = cls.__new_old__(cls, *args, **kw)
        it.__init_old__(*args, **kw)
        return it

    cls.__new__ = singleton_new
    cls.__init_old__ = cls.__init__
    cls.__init__ = object.__init__

    return cls



__all__ = [
    'typednamedtuple',
    'Heap',
    'TimeBasedHeap',
    'OmniHashableMixin',
    'Singleton',
    'DictObject',
    'apply_dict_object',
    'Immutable'
]

from .dict_object import *
from .singleton import Singleton
from .structures import *
from .typednamedtuple import *
from .immutable import Immutable



import ctypes
import logging
import platform
import threading

logger = logging.getLogger(__name__)


class TerminableThread(threading.Thread):
    """
    Class that will execute something in a loop unless terminated. Use like:

    >>> class MeGrimlock(TerminableThread):
    >>>     def loop(self):
    >>>         ... do your operations ..
    >>> a = MeGrimlock()
    >>> a.start()
    >>> a.terminate().join()

    Flag whether to terminate is stored in self._terminating.

    If you decide to override run(), you got to check periodically for self._terminating to become true.
    If you decide to use the loop/cleanup interface, you don't need to do so.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._terminating = False

    def loop(self) -> None:
        """
        Run one iteration of the loop. Meant to be overrided.

        This should block for as long as a single check will take, as termination checks take place
        between calls.
        """
        raise NotImplementedError('Override me!')

    def run(self) -> None:
        """
        Calls self.loop() indefinitely, until terminating condition is met
        """
        while not self._terminating:
            self.loop()
        self.cleanup()

    def cleanup(self):
        """
        Called after thread non-forced termination, in the thread's context.

        The default implementation does nothing.
        """

    def terminate(self, force: bool = False) -> 'TerminableThread':
        """
        Signal this thread to terminate.

        Forcing, if requested, will be done by injecting a SystemExit exception into target
        thread, so the thread must acquire GIL. For example, following would not be interruptable:

        Note that calling force=True on PyPy won't work, and RuntimeError will be raised instead.

        >>> time.sleep(1000000)

        :param force: Whether to force a quit
        :return: self
        :raises RuntimeError: when something goes wrong with the underlying Python machinery
        """
        self._terminating = True
        if force:
            if platform.python_implementation() == 'PyPy':
                raise RuntimeError('force=True was made on PyPy')
            ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), ctypes.py_object(SystemExit))
            if ret == 0:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Multiple threads killed!')
            elif ret > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Multiple threads killed!')

        return self



import types
import typing as tp
from abc import abstractmethod

__all__ = [
    'BaseExceptionHandler',
    'FunctionExceptionHandler',
    'exception_handler',
    'ALWAYS_FIRST',
    'ALWAYS_LAST',
    'NORMAL_PRIORITY',
    'ExceptionHandlerCallable'
]

ALWAYS_FIRST = -1000
NORMAL_PRIORITY = 0
ALWAYS_LAST = 1000

ExceptionHandlerCallable = tp.Callable[[type, BaseException, types.TracebackType], bool]


class BaseExceptionHandler:
    def __init__(self, priority=NORMAL_PRIORITY):
        """
        Instantiate an exception handler with provided priority.
        Handlers with smaller priorities run sooner.

        :param priority: Priority to use for this handler
        """
        self.priority = priority

    def install(self) -> 'BaseExceptionHandler':
        """
        Register this handler to run upon exceptions
        """
        from .global_eh import GlobalExcepthook
        GlobalExcepthook().add_hook(self)
        return self

    def uninstall(self):
        """
        Unregister this handler to run on exceptions
        """
        from .global_eh import GlobalExcepthook
        GlobalExcepthook().remove_hook(self)

    @abstractmethod
    def handle_exception(self, type_: tp.Callable[[type, BaseException, types.TracebackType], None], value,
                         traceback: types.TracebackType) -> tp.Optional[bool]:
        """
        Return True to intercept the exception, so that it won't be propagated to other handlers.
        """
        pass


class FunctionExceptionHandler(BaseExceptionHandler):
    """
    A exception handler to make callables of given signature into Satella's exception handlers.

    Your exception handler must return a bool, whether to intercept the exception and not propagate it.
    """

    def __init__(self, fun: ExceptionHandlerCallable, priority: int = NORMAL_PRIORITY):
        super(FunctionExceptionHandler, self).__init__(priority)
        self.fun = fun

    def handle_exception(self, type_, value, traceback):
        if type_ == SystemExit:
            return
        return self.fun(type_, value, traceback)


def exception_handler(priority: int = NORMAL_PRIORITY):
    """
    Convert a callable to an FunctionExceptionHandler. Usage

    >>> @exception_handler(priority=-10)
    >>> def handle_exc(type_, val, traceback):
    >>>     ...

    :return: ExceptionHandler instance
    """

    if not isinstance(priority, int):
        raise TypeError('Did you forget to use it as @exception_handler() ?')

    def outer(fun: ExceptionHandlerCallable) -> FunctionExceptionHandler:
        return FunctionExceptionHandler(fun, priority=priority)

    return outer



"""
POSIX things
"""

import os
import sys

from .daemon import daemonize
from .pidlock import PIDFileLock, LockIsHeld
from .signals import hang_until_sig

__all__ = [
    'daemonize',
    'PIDFileLock', 'LockIsHeld',
    'hang_until_sig',
    'is_running_as_root',
    'suicide'
]


def is_running_as_root() -> bool:
    """
    Is this process running as root?

    Checks whether EUID is 0

    :return: bool
    :raises OSError: called on Windows!
    """

    if sys.platform.startswith('win'):
        raise OSError('Routine unavailable on Windows!')

    return os.geteuid() == 0


def suicide(kill_entire_pg: bool = True) -> None:
    """
    Kill self.

    :param kill_entire_pg: whether to kill entire PG if a session leader
    :raises NotImplementedError: called on Windows!
    """

    if sys.platform.startswith('win'):
        raise NotImplementedError('suicide() not yet supported on Windows! (see issue #37)')

    mypid = os.getpid()
    kill = os.killpg if kill_entire_pg and os.getpgid(0) == mypid else os.kill
    kill(mypid, 9)



import typing as tp

from .base import Metric


class SimpleMetric(Metric):
    CLASS_NAME = 'string'
    CONSTRUCTOR = str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = None

    def append_child(self, metric: 'Metric'):
        raise TypeError('This metric cannot contain children!')

    def handle(self, level, data):
        if self.can_process_this_level(level):
            self.data = self.CONSTRUCTOR(data)

    def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
        return self.data


class StringMetric(SimpleMetric):
    pass


class IntegerMetric(SimpleMetric):
    CLASS_NAME = 'int'
    CONSTRUCTOR = int


class FloatMetric(SimpleMetric):
    CLASS_NAME = 'float'
    CONSTRUCTOR = float



from .base import Metric
from .cps import ClicksPerTimeUnitMetric
from .simple import StringMetric, IntegerMetric, FloatMetric

ALL_METRICS = [
    Metric,
    StringMetric,
    IntegerMetric,
    FloatMetric,
    ClicksPerTimeUnitMetric
]

METRIC_NAMES_TO_CLASSES = {
    metric.CLASS_NAME: metric for metric in ALL_METRICS
}



import typing as tp

from satella.json import JSONAble

DISABLED = 1
RUNTIME = 2
DEBUG = 3
INHERIT = 4


class Metric(JSONAble):
    """
    A base metric class
    """
    CLASS_NAME = 'base'

    def reset(self) -> None:
        """Delete all child metrics that this metric contains"""
        from satella.instrumentation import metrics
        if self.name == '':
            metrics.metrics = {}
        else:
            metrics.metrics = {k: v for k, v in metrics.metrics.items() if not k.startswith(self.name + '.')}
            del metrics.metrics[self.name]
        self.children = []

    def __init__(self, name, root_metric: 'Metric' = None, metric_level: str = None, **kwargs):
        """When reimplementing the method, remember to pass kwargs here!"""
        self.name = name
        self.root_metric = root_metric
        self.level = metric_level or RUNTIME
        assert not (self.name == '' and self.level == INHERIT), 'Unable to set INHERIT for root metric!'
        self.children = []

    def __str__(self) -> str:
        return self.name

    def append_child(self, metric: 'Metric'):
        self.children.append(metric)

    def can_process_this_level(self, target_level: int) -> bool:
        metric = self
        while metric.level == INHERIT:
            # this is bound to terminate, since it is not possible to set metric_level of INHERIT on root
            metric = metric.root_metric
        return metric.level >= target_level

    def switch_level(self, level: int) -> None:
        assert not (self.name == '' and level == INHERIT), 'Unable to set INHERIT for root metric!'
        self.level = level

    def to_json(self) -> tp.Union[list, dict, str, int, float, None]:
        return {
            child.name[len(self.name) + 1 if len(self.name) > 0 else 0:]: child.to_json() for child in self.children
        }

    def handle(self, level: int, *args, **kwargs) -> None:
        """Override me!"""
        raise TypeError('A collection of metrics is not meant to get .handle() called!')

    def debug(self, *args, **kwargs):
        self.handle(DEBUG, *args, **kwargs)

    def runtime(self, *args, **kwargs):
        self.handle(RUNTIME, *args, **kwargs)



import itertools
import logging
import threading
import typing as tp

logger = logging.getLogger(__name__)
from .metric_types.base import RUNTIME, DISABLED, DEBUG, INHERIT, Metric
from .metric_types import METRIC_NAMES_TO_CLASSES

__all__ = ['getMetric', 'DISABLED', 'RUNTIME', 'DEBUG', 'INHERIT']

metrics = {}
metrics_lock = threading.Lock()


def getMetric(metric_name: str, metric_type: str = 'base', metric_level: tp.Optional[str] = None, **kwargs):
    """
    Obtain a metric of given name.

    metric_name --  must be a module name
    """
    metric_level_to_set_for_children = metric_level or INHERIT
    name = metric_name.split('.')
    with metrics_lock:
        root_metric = None
        for name_index, name_part in itertools.chain(enumerate(name), ((len(name), None),)):
            tentative_name = '.'.join(name[:name_index])
            if tentative_name not in metrics:
                if tentative_name == '':
                    # initialize the root metric
                    if metric_level is None:
                        metric_level_to_set_for_root = RUNTIME
                    elif metric_level_to_set_for_children == INHERIT:
                        metric_level_to_set_for_root = RUNTIME
                    else:
                        metric_level_to_set_for_root = metric_level_to_set_for_children
                    print(metric_level_to_set_for_root)
                    metric = Metric('', None, metric_level_to_set_for_root)
                    metric.level = RUNTIME
                    root_metric = metric
                elif metric_name == tentative_name:
                    metric = METRIC_NAMES_TO_CLASSES[metric_type](tentative_name, root_metric, metric_level, **kwargs)
                else:
                    metric = Metric(tentative_name, root_metric, metric_level_to_set_for_children)
                metrics[tentative_name] = metric
                if metric != root_metric:  # prevent infinite recursion errors
                    root_metric.append_child(metric)
            else:
                metric = metrics[tentative_name]
            root_metric = metric

        if metric_level is not None:
            metric.switch_level(metric_level)

        return metric



# coding=UTF-8
"""
All things related to:
* post-mortem reporting
* on-line reporting
"""
from __future__ import print_function, absolute_import, division

from .dump_frames_on import install_dump_frames_on
from .trace_back import Traceback, StoredVariableValue, StackFrame, GenerationPolicy

__all__ = ['Traceback', 'StoredVariableValue', 'StackFrame', 'GenerationPolicy',
           'install_dump_frames_on']



def static_var(var_name: str, starting_value):
    """
    Declare a static variable for given function

    Use it like:

    >>> @static_var('counter', 2)
    >>> def count():
    >>>     count.counter += 1

    or:

    >>> class MyClass:
    >>>     @static_var('counter', 2)
    >>>     def count(self):
    >>>         MyClass.count.counter += 1
    """

    def decorate(func):
        setattr(func, var_name, starting_value)
        return func

    return decorate



class BaseSatellaException(Exception):
    def __init__(self, msg):
        super(BaseSatellaException, self).__init__()
        self.msg = msg

    def __str__(self):
        return 'BaseSatellaException(%s)' % (repr(self.msg),)


class ResourceLocked(BaseSatellaException):
    """Given resource has been already locked"""


class ResourceNotLocked(BaseSatellaException):
    """Locking given resource is needed in order to access it"""


class PreconditionError(BaseSatellaException, ValueError):
    """
    A precondition was not met for the argument
    """


class ConfigurationError(BaseSatellaException):
    """A generic error during configuration"""


class ConfigurationSchemaError(ConfigurationError):
    """Schema mismatch to what was seen"""


class ConfigurationValidationError(ConfigurationSchemaError):
    """A validator failed"""

    def __init__(self, msg, value=None):
        """
        :param value: value found
        """
        super(ConfigurationValidationError, self).__init__(msg)
        self.value = value



from .callablegroup import *
from .locked_dataset import LockedDataset
from .monitor import *
from .thread import TerminableThread

__all__ = ['LockedDataset', 'Monitor', 'RMonitor', 'CallableGroup', 'TerminableThread']



import functools
import typing as tp

__all__ = [
    'rethrow_as',
    'silence_excs'
]


def silence_excs(*exc_types):
    """
    Silence given exception types.

    Can be either a decorator or a context manager
    """
    return rethrow_as(exc_types, None)


class rethrow_as:
    """
    Transform some exceptions into others.

    Either a decorator or a context manager

    New exception will be created by calling exception to transform to with
    repr of current one.

    You can also provide just two exceptions, eg.

    >>> rethrow_as(NameError, ValueError)

    You can also provide a pairwise translation, eg. from NameError to ValueError and from OSError to IOError

    >>> rethrow_as((NameError, ValueError), (OSError, IOError))

    If the second value is a None, exception will be silenced.
    """

    def __init__(self, *pairs, exception_preprocessor: tp.Optional[tp.Callable[[Exception], str]] = None):
        """
        Pass tuples of (exception to catch - exception to transform to).

        :param exception_preprocessor: other callable/1 to use instead of repr.
            Should return a str, a text description of the exception
        """
        try:
            a, b = pairs  # throws ValueError
            op = issubclass(b, BaseException)  # throws TypeError
        except TypeError:
            op = b is None
        except ValueError:
            op = False

        if op:
            pairs = [pairs]

        self.mapping = list(pairs)
        self.exception_preprocessor = exception_preprocessor or repr

    def __call__(self, fun: tp.Callable) -> tp.Any:
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            with self:
                return fun(*args, **kwargs)

        return inner

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            for from_, to in self.mapping:
                if issubclass(exc_type, from_):
                    if to is None:
                        return True
                    else:
                        raise to(self.exception_preprocessor(exc_val))



from . import base
from . import derivative
from . import envvars
from . import file
from . import format
from . import from_dict
from .base import *
from .derivative import *
from .envvars import *
from .file import *
from .format import *
from .from_dict import *

__all__ = format.__all__ + envvars.__all__ + derivative.__all__ + file.__all__ + base.__all__ + from_dict.__all__



import collections
import time
import typing as tp

from .base import Metric


class ClicksPerTimeUnitMetric(Metric):
    """
    This tracks the amount of calls to handle() during the last time periods, as specified by time_unit_vectors
    (in seconds). You may specify multiple time periods as consequent entries in the list.
    """
    CLASS_NAME = 'cps'

    def __init__(self, *args, time_unit_vectors: tp.Optional[tp.List[float]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        time_unit_vectors = time_unit_vectors or [1]
        self.last_clicks = collections.deque()
        self.cutoff_period = max(time_unit_vectors)
        self.time_unit_vectors = time_unit_vectors

    def handle(self, level: int, *args, **kwargs) -> None:
        monotime = time.monotonic()
        if self.can_process_this_level(level):
            self.last_clicks.append(time.monotonic())
            try:
                while self.last_clicks[0] <= monotime - self.cutoff_period:
                    self.last_clicks.popleft()
            except IndexError:
                pass

    def to_json(self) -> tp.List[int]:
        count_map = [0] * len(self.time_unit_vectors)
        monotime = time.monotonic()
        time_unit_vectors = [monotime - v for v in self.time_unit_vectors]

        for v in self.last_clicks:
            for index, cutoff in enumerate(time_unit_vectors):
                if v >= cutoff:
                    count_map[index] += 1

        return count_map



import sys
import time
import typing as tp

from satella.posix import suicide
from .exception_handlers import BaseExceptionHandler, ALWAYS_FIRST, ExceptionHandlerCallable


class MemoryErrorExceptionHandler(BaseExceptionHandler):
    """
    A handler that terminates the entire process (or process group) is a MemoryError is seen.

    `custom_hook` is an exception callable to implement you own behavior. If it returns True,
    then MemoryErrorExceptionHandler won't kill anyone.
    """

    def __init__(self, custom_hook: ExceptionHandlerCallable = lambda type_, value, traceback: False,
                 kill_pg: bool = False):
        """
        :param kill_pg: whether to kill entire process group, if applicable
        """
        super(MemoryErrorExceptionHandler, self).__init__()
        self.priority = ALWAYS_FIRST  # always run first!
        # so that we have some spare space in case a MemoryError is thrown
        self._free_on_memory_error = {'a': bytearray(1024 * 2)}
        self.custom_hook = custom_hook
        self.kill_pg = kill_pg
        self.installed = False

    def install(self):
        if self.installed:
            raise RuntimeError('already installed')

        from .global_eh import GlobalExcepthook
        GlobalExcepthook().add_hook(self)

    def handle_exception(self, type_, value, traceback) -> tp.Optional[bool]:
        if not issubclass(type_, MemoryError):
            return

        del self._free_on_memory_error['a']

        # noinspection PyBroadException
        try:
            if self.custom_hook(type_, value, traceback):
                return True
        except Exception as e:
            pass

        try:
            sys.stderr.write(
                'satella.exception_handling: MemoryError seen, killing\n')
            sys.stderr.flush()
        except (IOError, OSError):
            pass

        suicide(self.kill_pg)

        time.sleep(5)
        return True



import typing as tp

__all__ = ['DictObject', 'apply_dict_object']


class DictObject(dict):
    def __getattr__(self, item: str) -> tp.Any:
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(repr(e))

    def __setattr__(self, key: str, value: tp.Any) -> None:
        self[key] = value

    def __delattr__(self, key: str) -> None:
        try:
            del self[key]
        except KeyError as e:
            raise AttributeError(repr(e))


def apply_dict_object(v: tp.Any) -> tp.Union[DictObject, tp.Any]:
    """
    Apply DictObject() to every dict inside v.

    This assumes that the only things that will be touched will be nested dicts and lists.

    If you pass a non-dict and a non-list, they will be returned as is.
    """
    if isinstance(v, list):
        return [apply_dict_object(x) for x in v]
    elif isinstance(v, dict):
        return DictObject({
            k: apply_dict_object(val) for k, val in v.items()
        })
    else:
        return v



import abc

from .metric_types.base import RUNTIME, INHERIT


class BaseRecord(metaclass=abc.ABCMeta):
    def __init__(self, level: str = RUNTIME):
        self.level = level

    def can_be_handled_by(self, metric: 'Metric'):
        run_level = metric.level
        while run_level == INHERIT:
            metric = metric.root_metric
            run_level = metric.level
        else:
            return run_level >= self.level
        raise ValueError('Invalid metric setup - root metric is in inherit mode!')



import functools
import itertools

from ..exceptions import PreconditionError

__all__ = ['precondition', 'for_argument', 'PreconditionError']

_NOP = lambda x: x
_TRUE = lambda x: True


def precondition(*t_ops):
    """
    Check that a precondition happens for given parameter.
    Only positional arguments are supported.

    You can do it like this:

    >>> @precondition(lambda x: x == 1)
    >>> def return_two(x):
    >>>     return x*2

    or

    >>> @precondition('x == 1')
    >>> def return_two(x):
    >>>     ..

    If None is passed then argument will be always assumed to be True.
    You can use all standard locals in precondition.

    You function call will return a PreconditionError (subclass of
    ValueError) if a precondition fails
    """

    tn_ops = []

    for t_op in t_ops:
        if t_op is None:
            precond = _TRUE
        elif isinstance(t_op, str):
            q = dict(globals())
            exec('_precond = lambda x: ' + t_op, q)
            precond = q['_precond']
        else:
            precond = t_op

        tn_ops.append(precond)

    from satella.coding.recast_exceptions import rethrow_as

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            assert len(args) >= len(tn_ops), 'More preconditions than positional arguments!'
            with rethrow_as(TypeError, PreconditionError):
                for arg, precond in itertools.zip_longest(args, tn_ops, fillvalue=_TRUE):
                    if not precond(arg):
                        raise PreconditionError(
                            'Argument of value %s failed precondition check' % (arg,))
            return fun(*args, **kwargs)

        return inner

    return outer


def for_argument(*t_ops, **t_kwops):
    """
    Calls a callable for each of the arguments. Pass None if you do not wish to process given argument.

    returns is a special keyword, a callable to process the result through

    Use like:

    >>> @for_argument(int, str, typed=bool, returns=int)
    >>> def check(val1, val2, typed='True'):
    >>>     if typed:
    >>>         return val1 + int(val2)
    """
    t_ops = [_NOP if op == 'self' else op for op in t_ops]
    returns = t_kwops.pop('returns', _NOP)

    def outer(fun):
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            # add extra 'None' argument if unbound method
            assert len(args) >= len(t_ops)
            a = fun(*((_NOP if op is None else op)(arg) for arg, op in
                      itertools.zip_longest(args, t_ops, fillvalue=None)),
                    **{k: t_kwops.get(k, _NOP)(v) for k, v in kwargs.items()})
            return returns(a)

        return inner

    return outer
